/* ------------------------- line.c ----------------------------------------- */

/* Draw lines on SVGA. Can be very fast when usage of low level assembly
 * support is requested. It handles VGA page crossing.
 *
 * It originated from the DJgpp package but completely re-written since, the
 * only surviving clue being the WriteMode specification in the color field.
 * Actually this is mostly not used either.
 *
 * GrMoveTo (x, y)
 * GrDrawTo (x, y, color)
 * GrLine (x0, y0, x1, y1, color)
 * GrClear (x, y, sx, sy, color)
 * GrPolygon (npoints, points, color)
 *
 * These functions use some asm support from subs.asm:
 *	_GrAddr
 *	_ILoop
 *	_SLoop
 *	vgaset
 *
 * If you only use the low-level *A() asm version then this module is
 * redundant.
*/

#include <stdlib.h>
#include <string.h>

#ifndef USE_ASM
#define USE_ASM
#endif
#include "grh.h"


#define FAST_PLOT(va) \
	if (!_GrWriteMode)			\
		*va = (uchar)c;			\
	else if (_GrWriteMode & GrXOR)		\
		*va ^= c;			\
	else /* if (_GrWriteMode & GrXOR) */	\
		*va |= c

extern int far _ILoop (uint iva, int dy, int dx, int dvx, int dvy, int c);
#define _InnerLoop	_ILoop

extern int far _SLoop (uint va, int count, int sv, int c);
#define _SimpleLoop	_SLoop


static uint	x1 = 0, y1 = 0, xydone = 0;
static uchar	far *va = VGA_PAGE;


extern void far
GrMoveTo (uint x, uint y)
{
#if 0
	if (x >= _GrSizeX)
		x = _GrSizeX-1;
	if (y >= _GrSizeY)
		y = _GrSizeY-1;
#endif
	x1 = x;
	y1 = y;
	xydone = 0;
}

extern void far
GrDrawTo (uint x2, uint y2, uint c)
{
	register int	dx, dy, svx, svy;
	int		t;

#if 0
	if (x2 >= _GrSizeX)
		x2 = _GrSizeX-1;
	if (y2 >= _GrSizeY)
		y2 = _GrSizeY-1;
#endif
	t = 0;
	if (!xydone) {
		t = 1;
		xydone = 1;
		MY_OFF(va) = (ushort)_GrAddr (x1, y1);
		FAST_PLOT(va);
	}

	svx = 1;
	if ((dx = x2 - x1) < 0) {
		dx = -dx;
		svx = -svx;
	}

	svy = _GrSizeX;
	if ((dy = y2 - y1) < 0) {
		dy = -dy;
		svy = -svy;
	}

	if (dx > dy) {
		if (0 == dy) {
			int	Page;

			Page = _WritePage;
			c |= _GrWriteMode;
			if (svx >= 0) {
				ushort	room;

				room = 0xffffU - MY_OFF(va);
				if ((uint)dx <= room) {
					vgaset (MY_OFF(va)+1, c, dx);
					MY_OFF(va) += dx;
				} else {
					if (room > 0)
						vgaset (MY_OFF(va)+1, c, room);
					++Page, _GrSetPage (Page, Page);
					MY_OFF(va) = dx - room - 1;
					vgaset (0, c, MY_OFF(va)+1);
				}
			} else {
				if (MY_OFF(va) >= (uint)dx) {
					MY_OFF(va) -= dx;
					vgaset (MY_OFF(va), c, dx);
				} else {
					vgaset (0, c, MY_OFF(va));
					--Page, _GrSetPage (Page, Page);
					dx -= MY_OFF(va);
					MY_OFF(va) = (ushort)(0x10000UL-dx);
					vgaset (MY_OFF(va), c, dx);
				}
			}
		} else
			MY_OFF(va) = _InnerLoop (MY_OFF(va), dy, dx, svx, svy,
					c);
		++GrStats[dx+t];
	} else if (dx < dy) {
		if (0 == dx)
			MY_OFF(va) = _SimpleLoop (MY_OFF(va), dy, svy, c);
		else
			MY_OFF(va) = _InnerLoop (MY_OFF(va), dx, dy, svy, svx,
					c);
		++GrStats[dy+t];
	} else /* if (dx == dy) */ {
		if (dx)
			MY_OFF(va) = _SimpleLoop (MY_OFF(va), dx, svx+svy, c);
		++GrStats[dx+t];
	}

	x1 = x2;
	y1 = y2;
}

extern void far
GrLine (uint xa, uint ya, uint xb, uint yb, uint c)
{
	if (xa != x1 || ya != y1 || !xydone)
		GrMoveTo (xa, ya);
#if 0
	if ((c&0xff00) != _GrWriteMode)
		GrSetWriteMode (c);
#endif
	GrDrawTo (xb, yb, c);
}

extern void far
GrClear (uint x, uint y, uint sx, uint sy, uint color)
{
	long	room;
	int	Page;

	if (!sx || !sy)
		return;

	if (!_GrWriteMode) {
		MY_OFF(va) = (ushort)_GrAddr (x, y);
		room = 0x10000UL - MY_OFF(va);	/* room to end of bank */
		Page = _WritePage;
		for (; sy--; ++y) {
			if (room <= 0) {
				room += 0x10000UL;
				++Page, _GrSetPage (Page, Page);
			}
			if ((long)sx > room) {
				vgaset (MY_OFF(va), color, (uint)room);
				++Page, _GrSetPage (Page, Page);
				vgaset (0, color, sx-(uint)room);
				MY_OFF(va) = _GrSizeX - (uint)room;
				room = 0x10000UL - MY_OFF(va);
			} else {
				vgaset (MY_OFF(va), color, sx);
				MY_OFF(va) += _GrSizeX;
				room -= _GrSizeX;
			}
		}
		GrMoveTo (x+sx-1, y-1);
	} else {
		for (sx += x-1; sy--; ++y) {
			GrMoveTo (x,  y);
			GrDrawTo (sx, y, color);
		}
	}
}

extern void far
GrPolygon (int npoints, ushort *points, uint color)
{
	uint	y;			/* current y */
	int	fwdx, fwdy, bwdx, bwdy;	/* current edge tops */
	int	fdx, fdy, bdx, bdy;	/* interpolation */
	int	fx, bx;			/* segment endpoints */
	ushort	*lastpoint;		/* end of 'points' */
	ushort	*fwd, *bwd;		/* current point */
	ushort	*nfwd, *nbwd;		/* next point */
	long	room;
	uint	n;
	int	Page;

	if (npoints < 2)
		return;

	if (2 == npoints) {
		GrMoveTo (points[1], points[2]);
		GrDrawTo (points[3], points[4], color);
		return;
	}

/* locate the top vertex (use fwd as temp).
*/
	lastpoint = points + 2*(npoints - 1);
	y = (fwd = points)[1];
	for (bwd = points+2; bwd <= lastpoint; bwd += 2) {
		if (y > bwd[1])
			y = (fwd = bwd)[1];
	}

/* run through the edges, top to bottom.
*/
	nfwd = nbwd = fwd;

	if (!_GrWriteMode) {
		MY_OFF(va) = (ushort)_GrAddr (0, y);
		room = 0x10000UL - MY_OFF(va);	/* room to end of bank */
		Page = _WritePage;
	}

	for (;;) {
		if (y == nfwd[1]) {
			do {
				fwd = nfwd;
				if (fwd == lastpoint)
					nfwd = points;
				else
					nfwd = fwd + 2;
			} while (y == nfwd[1]);
			if ((fdy = nfwd[1] - fwd[1]) < 0)
				break;
			fdx = nfwd[0] - (fwdx = fwd[0]);
			fwdy = 0;
		}

		if (y == nbwd[1]) {
			do {
				bwd = nbwd;
				if (bwd == points)
					nbwd = lastpoint;
				else
					nbwd = bwd - 2;
			} while (y == nbwd[1]);
			if ((bdy = nbwd[1] - bwd[1]) < 0)
				break;
			bdx = nbwd[0] - (bwdx = bwd[0]);
			bwdy = 0;
		}

		fx = fwdx + _muldiv (fwdy, fdx, fdy);
		bx = bwdx + _muldiv (bwdy, bdx, bdy);
		if (!_GrWriteMode) {
			if (fx > bx) {
				n  = fx;
				fx = bx;
				bx = n;
			}
			n = bx - fx + 1;
			if ((long)fx >= room)
				++Page, _GrSetPage (Page, Page);
			MY_OFF(va) = (uint)(0x0ffffU & (MY_OFF(va) + fx));
			room = 0x10000UL - MY_OFF(va);
			if ((long)n > room) {
				vgaset (MY_OFF(va), color, (uint)room);
				n -= (uint)room;
				MY_OFF(va) = 0;
				room = 0x10000UL;
				++Page, _GrSetPage (Page, Page);
			}
			vgaset (MY_OFF(va), color, n);
			n += _GrSizeX - bx - 1;
			MY_OFF(va) += n;
			room -= n;
		} else {
			GrMoveTo (fx, y);
			GrDrawTo (bx, y, color);
		}
		++y;
		++fwdy;
		++bwdy;
	}

/* do bottom tip/edge.
*/
	fx = fwdx + fdx;
	bx = bwdx + bdx;
	if (!_GrWriteMode) {
		if (fx > bx) {
			n  = fx;
			fx = bx;
			bx = n;
		}
		n = bx - fx + 1;
		if ((long)fx >= room)
			++Page, _GrSetPage (Page, Page);
		MY_OFF(va) = (uint)(0x0ffffU & (MY_OFF(va) + fx));
		room = 0x10000UL - MY_OFF(va);
		if ((long)n > room) {
			vgaset (MY_OFF(va), color, (uint)room);
			n -= (uint)room;
			MY_OFF(va) = 0;
			++Page, _GrSetPage (Page, Page);
		}
		vgaset (MY_OFF(va), color, n);
		GrMoveTo (bx, y);	/* restore currency */
	} else {
		GrMoveTo (fx, y);
		GrDrawTo (bx, y, color);
	}
}
